// © 2023 John Breauxs
use crate::parser::{error::ParseError, preamble::*};
use std::{
    fmt::Display,
    path::{Path, PathBuf},
};

#[derive(Debug)]
pub enum AssemblyError {
    UnresolvedIdentifier(Identifier),
    RedefinedLabel(Identifier),
    JumpedTooFar(Identifier, isize),
    ParseError(ParseError),
    // TODO: This, better'
    Context(Box<AssemblyError>, PathBuf, usize),
    EmptyBuffer,
}

impl AssemblyError {
    pub(super) fn ctx<P: AsRef<Path> + ?Sized>(self, file: &P, line: usize) -> Self {
        Self::Context(self.into(), file.as_ref().into(), line)
    }
}

impl From<ParseError> for AssemblyError {
    fn from(value: ParseError) -> Self { Self::ParseError(value) }
}

impl Display for AssemblyError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::UnresolvedIdentifier(id) => {
                write!(f, "Identifier {id} is undefined, but referenced anyway.")
            }
            Self::RedefinedLabel(id) => {
                write!(f, "Redefined label '{id}'.")
            }
            Self::JumpedTooFar(id, num) => {
                write!(f, "Label '{id}' is too far away. ({num} is outside range -0x400..=0x3fe)")
            }
            Self::ParseError(e) => Display::fmt(e, f),
            Self::Context(e, file, line) => write!(f, "{}:{line}:\n\t{e}", file.display()),
            Self::EmptyBuffer => Display::fmt("Tried to get last element of output buffer, but buffer was empty", f),
        }
    }
}

impl std::error::Error for AssemblyError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::ParseError(e) => Some(e),
            Self::Context(e, ..) => Some(e),
            _ => None,
        }
    }
}
