// © 2023 John Breauxs
use super::*;
use crate::lexer::error::LexError;

#[derive(Debug)]
pub enum ParseError {
    /// Produced by [lexer](crate::lexer)
    LexError(LexError),
    /// Produced by [std::io]
    IoError(std::io::Error),
    /// Produced by [Number](Number)[::parse()](Parsable::parse())
    /// when the parsed number contains digits too high for the specified radix
    UnexpectedDigits(String, u32),
    /// Produced by [Opcode](Opcode)[::parse()](Parsable::parse())
    /// when the opcode passed lexing but did not match recognized opcodes.
    ///
    /// This is always a lexer bug.
    UnrecognizedOpcode(String),
    /// Produced by [Directive](Directive)[::parse()](Parsable::parse())
    /// when an unknown or unimplemented directive is used
    UnrecognizedDirective(String),
    /// Produced by [Register] when attempting to convert from a [str]
    /// that isn't a register (pc, sp, sr, cg, or r{number})
    NotARegister(String),
    /// Produced by [Register] when the r{number} is outside the range 0-15
    RegisterTooHigh(u16),
    /// Produced by [SecondaryOperand] when the joke "secondary immediate" form
    /// is out of range 0..=1
    FatSecondaryImmediate(isize),
    /// Produced by a [Number] too wide to fit in 16 bits
    /// (outside the range `(-2^15) .. (2^16-1)` )
    NumberTooWide(isize),
    /// Produced by [JumpTarget](parser::preamble::JumpTarget)
    /// when the jump offset is outside the range (-0x3ff..0x3fc)
    JumpedTooFar(isize),
    /// Produced by [JumpTarget](parser::preamble::JumpTarget)
    JumpedOdd(isize),
}

impl From<LexError> for ParseError {
    fn from(value: LexError) -> Self { Self::LexError(value) }
}
impl From<std::io::Error> for ParseError {
    fn from(value: std::io::Error) -> Self { Self::IoError(value) }
}

impl Display for ParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::LexError(error) => Display::fmt(error, f),
            Self::IoError(error) => Display::fmt(error, f),
            Self::UnexpectedDigits(number, radix) => write!(f, "Number `{number}` is not base {radix}."),
            Self::UnrecognizedOpcode(op) => write!(f, "{op} is not an opcode"),
            Self::UnrecognizedDirective(d) => write!(f, "{d} is not a directive."),
            Self::NotARegister(reg) => write!(f, "{reg} is not a register"),
            Self::RegisterTooHigh(reg) => write!(f, "r{reg} is not a register"),
            Self::FatSecondaryImmediate(num) => write!(f, "Secondary immediate must be #0 or #1, not #{num}"),
            Self::NumberTooWide(num) => write!(f, "{num} does not fit in 16 bits"),
            Self::JumpedTooFar(num) => write!(f, "{num} is too far away: must be in range (`-1022..=1024`.)"),
            Self::JumpedOdd(num) => {
                write!(f, "Jump targets only encode even numbers: {num} must not be odd.")
            }
        }
    }
}
impl std::error::Error for ParseError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::LexError(e) => Some(e),
            Self::IoError(e) => Some(e),
            _ => None,
        }
    }
}
