// © 2023 John Breaux
//! [`Line`] contains a single subcomponent of the document. Multiple instructions on the same
//! document line will be treated as if they took up multiple [`Line`s](Line).
//!
//! A line contains one of:
//! - [`Label`]
//! - [`Instruction`]
//! - [`Directive`]
//! - [`Comment`]
//! - [Nothing](Line::Empty)
use super::*;

/// A line contains any one of:
/// - [`Label`] (definition)
/// - [`Instruction`]
/// - [`Directive`]
/// - [`Comment`]
/// - Nothing at all
#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Line {
    Empty,
    Insn(Instruction),
    Comment(Comment),
    Directive(Directive),
    Label(Label),
    EndOfFile, // Expected end of file
}

impl Parsable for Line {
    fn parse<'text, T>(p: &Parser, stream: &mut T) -> Result<Self, ParseError>
    where T: TokenStream<'text> {
        Ok(
            match stream
                .peek_expect_any_of([
                    Type::Endl,
                    Type::Insn,
                    Type::Comment,
                    Type::Directive,
                    Type::Identifier,
                    Type::EndOfFile,
                ])?
                .variant()
            {
                Type::Endl => {
                    stream.next();
                    Self::Empty
                }
                Type::Insn => Self::Insn(Instruction::parse(p, stream)?),
                Type::Comment => Self::Comment(Comment::parse(p, stream)?),
                Type::Directive => Self::Directive(Directive::parse(p, stream)?),
                Type::Identifier => Self::Label(Label::parse(p, stream)?),
                Type::EndOfFile => {
                    stream.next();
                    Self::EndOfFile
                }
                _ => unreachable!("stream.peek_expect_any_of should return Err for unmatched inputs"),
            },
        )
    }
}
impl Display for Line {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Empty => writeln!(f, "\n"),
            Self::Label(arg0) => Display::fmt(arg0, f),
            Self::Insn(arg0) => Display::fmt(arg0, f),
            Self::Directive(arg0) => Display::fmt(arg0, f),
            Self::Comment(arg0) => Display::fmt(arg0, f),
            Self::EndOfFile => write!(f, "; End of file."),
        }
    }
}
